from sklearn.metrics import pairwise
import numpy as np

from ...multiview.multiview_utils import BaseMultiviewClassifier, get_examples_views_indices
from ...utils.hyper_parameter_search import CustomUniform, CustomRandint
from ...utils.transformations import sign_labels, unsign_labels

class KernelClassifier(BaseMultiviewClassifier):

    def __init__(self, random_state=None,):
        super().__init__(random_state)

    def _compute_kernels(self, X, example_indices, view_indices, ):
        new_X = {}
        for index, (kernel_function, kernel_config, view_index) in enumerate(
                zip(self.kernel_functions, self.kernel_configs, view_indices)):
            new_X[index] = kernel_function(X.get_v(view_index,
                                                   example_indices),
                                           **kernel_config)
        return new_X

    def format_X(self, X, example_indices, view_indices):
        example_indices, view_indices = get_examples_views_indices(X,
                                                                   example_indices,
                                                                   view_indices)
        formatted_X = dict((index, X.get_v(view_index, example_indices=example_indices))
                     for index, view_index in enumerate(view_indices))

        return formatted_X, example_indices

    def extract_labels(self, predicted_labels):
        signed_labels = np.sign(predicted_labels)
        return unsign_labels(signed_labels)

    def init_kernels(self, nb_view=2, ):
        if isinstance(self.kernel, KernelDistribution):
            self.kernel = self.kernel.draw(nb_view)
        elif isinstance(self.kernel, str):
            self.kernel = [self.kernel
                                     for _ in range(nb_view)]
        elif isinstance(self.kernel, list):
            pass

        if isinstance(self.kernel_params, KernelConfigDistribution):
            self.kernel_params = self.kernel_params.draw(nb_view)
            self.kernel_params = [kernel_config[kernel_name]
                                   for kernel_config, kernel_name
                                   in zip(self.kernel_params,
                                          self.kernel)]

        elif isinstance(self.kernel_params, dict):
            self.kernel_params = [self.kernel_params for _ in range(nb_view)]
        else:
            pass


class KernelConfigGenerator:

    def __init__(self):
        pass

    def rvs(self, random_state=None):
        return KernelConfigDistribution(seed=random_state.randint(1))


class KernelConfigDistribution:

    def __init__(self, seed=42):
        self.random_state=np.random.RandomState(seed)
        self.possible_config = {
            "chi2": {"gamma": CustomUniform()},
            "rbf": {"gamma": CustomUniform()},
        }

    def draw(self, nb_view):
        drawn_params = [{} for _ in range(nb_view)]
        for view_index in range(nb_view):
            for kernel_name, params_dict in self.possible_config.items():
                drawn_params[view_index][kernel_name] = {}
                for param_name, distrib in params_dict.items():
                    drawn_params[view_index][kernel_name][param_name] = distrib.rvs(self.random_state)
        return drawn_params


class KernelGenerator:

    def __init__(self):
        pass

    def rvs(self, random_state=None):
        return KernelDistribution(seed=random_state.randint(1))


class KernelDistribution:

    def __init__(self, seed=42):
        self.random_state=np.random.RandomState(seed)
        self.available_kernels = ["rbf"]

    def draw(self, nb_view):
        return list(self.random_state.choice(self.available_kernels, nb_view))
